#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 17 03:28:27 2019

@author: carl
"""

from octavvs import clustering

if __name__ == '__main__':
    clustering.main()
    
    
    